From 42b31eef4ba92252be275371c24edc0172bd0d57 Mon Sep 17 00:00:00 2001
From: Sergey Radionov <RSATom@gmail.com>
Date: Tue, 30 Aug 2011 20:59:25 +0700
Subject: [PATCH 4/6] ActiveX: Changed manner of main thread event
 notification to winodow messages. This allows avoid
 deadlocks in some cases.

---
 projects/activex/connectioncontainer.cpp |  215 +++++++++++++++++++-----------
 projects/activex/connectioncontainer.h   |    7 +-
 2 files changed, 142 insertions(+), 80 deletions(-)

diff --git a/projects/activex/connectioncontainer.cpp b/projects/activex/connectioncontainer.cpp
index 91fd0b6..42c37e9 100644
--- a/projects/activex/connectioncontainer.cpp
+++ b/projects/activex/connectioncontainer.cpp
@@ -102,80 +102,150 @@ public:
 };
 
 ////////////////////////////////////////////////////////////////////////////////////////////////
-// Condition variable emulation
+// EventSystemProxyWnd
 ////////////////////////////////////////////////////////////////////////////////////////////////
-
-static void ConditionInit(HANDLE *handle)
+class EventSystemProxyWnd
 {
-    *handle = CreateEvent(NULL, TRUE, FALSE, NULL);
-    assert(*handle != NULL);
-}
+private:
+    enum{
+        WM_NEW_EVENT_NOTIFY = WM_USER+1
+    };
+    static LPCTSTR getClassName(void) { return TEXT("VLC ActiveX ESP Class"); };
+    static void RegisterWndClassName();
+    static void UnRegisterWndClassName();
+    static LRESULT CALLBACK ESPWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
+
+    static HINSTANCE _hInstance;
+    static ATOM _ESP_wndclass_atom;
+    HWND _hWnd;
+
+public:
+    static EventSystemProxyWnd* CreateESPWindow(HINSTANCE hInstance, VLCConnectionPointContainer *pCPC);
+
+private:
+    EventSystemProxyWnd(HWND hWnd, VLCConnectionPointContainer *pCPC)
+        :_pCPC(pCPC), _HasUnprocessedNotify(false), _hWnd(hWnd){};
+    void ProcessNotify();
+
+public:
+    void DestroyWindow()
+        {::DestroyWindow(_hWnd);};
+
+    //allowed to invoke from any thread
+    void NewEventNotify();
+
+private:
+    VLCConnectionPointContainer *_pCPC;
+    volatile bool _HasUnprocessedNotify;
+};
+
+HINSTANCE EventSystemProxyWnd::_hInstance=0;
+ATOM EventSystemProxyWnd::_ESP_wndclass_atom=0;
 
-static void ConditionDestroy(HANDLE *handle)
-{
-    CloseHandle(*handle);
-}
 
-static void ConditionWait(HANDLE *handle, CRITICAL_SECTION *lock)
+void EventSystemProxyWnd::RegisterWndClassName()
 {
-    DWORD dwWaitResult;
+    WNDCLASS wClass;
+    memset(&wClass, 0, sizeof(WNDCLASS));
 
-    do {
-        LeaveCriticalSection(lock);
-        dwWaitResult = WaitForSingleObjectEx(*handle, INFINITE, TRUE);
-        EnterCriticalSection(lock);
-    } while(dwWaitResult == WAIT_IO_COMPLETION);
+    if( ! GetClassInfo(_hInstance, getClassName(), &wClass) )
+    {
+        wClass.lpfnWndProc    = ESPWindowProc;
+        wClass.hInstance      = _hInstance;
+        wClass.lpszClassName  = getClassName();
 
-    assert(dwWaitResult != WAIT_ABANDONED); /* another thread failed to cleanup! */
-    assert(dwWaitResult != WAIT_FAILED);
-    ResetEvent(*handle);
+        _ESP_wndclass_atom = RegisterClass(&wClass);
+    }
+    else
+    {
+        _ESP_wndclass_atom = 0;
+    }
 }
 
-static void ConditionSignal(HANDLE *handle)
+void EventSystemProxyWnd::UnRegisterWndClassName()
 {
-    SetEvent(*handle);
+    if(0 != _ESP_wndclass_atom){
+        UnregisterClass(MAKEINTATOM(_ESP_wndclass_atom), _hInstance);
+        _ESP_wndclass_atom = 0;
+    }
 }
 
-////////////////////////////////////////////////////////////////////////////////////////////////
-// Event handling thread
-//
-// It bridges the gap between libvlc thread context and ActiveX/COM thread context.
-////////////////////////////////////////////////////////////////////////////////////////////////
-DWORD WINAPI ThreadProcEventHandler(LPVOID lpParam)
+LRESULT CALLBACK EventSystemProxyWnd::ESPWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
 {
-    CoInitialize(NULL);
-    VLCConnectionPointContainer *pCPC = (VLCConnectionPointContainer *)lpParam;
+    EventSystemProxyWnd* ESP = reinterpret_cast<EventSystemProxyWnd*>(GetWindowLongPtr(hWnd, GWLP_USERDATA));
 
-    while(pCPC->isRunning)
+    switch( uMsg )
     {
-        EnterCriticalSection(&(pCPC->csEvents));
-        ConditionWait(&(pCPC->sEvents), &(pCPC->csEvents));
+        case WM_CREATE:{
+            CREATESTRUCT* CreateStruct = (CREATESTRUCT*)(lParam);
+            VLCConnectionPointContainer * pCPC = reinterpret_cast<VLCConnectionPointContainer *>(CreateStruct->lpCreateParams);
 
-        if (!pCPC->isRunning)
-        {
-            LeaveCriticalSection(&(pCPC->csEvents));
+            ESP = new EventSystemProxyWnd(hWnd, pCPC);
+            SetWindowLongPtr(hWnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(ESP));
             break;
         }
+        case WM_NCDESTROY:
+            delete ESP;
+            SetWindowLongPtr(hWnd, GWLP_USERDATA, 0);
+            break;
+        case WM_NEW_EVENT_NOTIFY:
+            ESP->ProcessNotify();
+            break;
+        default:
+            return DefWindowProc(hWnd, uMsg, wParam, lParam);
+    }
+    return 0;
+}
 
-        while(!pCPC->_q_events.empty())
-        {
-            VLCDispatchEvent *ev = pCPC->_q_events.front();
-            pCPC->_q_events.pop();
-            pCPC->_p_events->fireEvent(ev->_dispId, &ev->_dispParams);
-            delete ev;
+EventSystemProxyWnd* EventSystemProxyWnd::CreateESPWindow(HINSTANCE hInstance, VLCConnectionPointContainer *pCPC)
+{
+    //save hInstance for future use
+    _hInstance = hInstance;
 
+    RegisterWndClassName();
 
-            if (!pCPC->isRunning)
-            {
-                LeaveCriticalSection(&(pCPC->csEvents));
-                goto out;
-            }
+    HWND hWnd = CreateWindow(getClassName(),
+                             0,
+                             0,
+                             0, 0, 0, 0,
+                             0,
+                             0,
+                             _hInstance,
+                             pCPC
+                             );
+
+    if(hWnd)
+        return reinterpret_cast<EventSystemProxyWnd*>(GetWindowLongPtr(hWnd, GWLP_USERDATA));
+
+    return 0;
+}
+
+void EventSystemProxyWnd::NewEventNotify()
+{
+    if(!_HasUnprocessedNotify){
+        _HasUnprocessedNotify=true;
+        PostMessage(_hWnd, WM_NEW_EVENT_NOTIFY, 0, 0);
+    }
+}
+
+void EventSystemProxyWnd::ProcessNotify()
+{
+    while(_pCPC->isRunning){
+        VLCDispatchEvent *ev = 0;
+        EnterCriticalSection(&(_pCPC->csEvents));
+        if(!_pCPC->_q_events.empty()){
+            ev = _pCPC->_q_events.front();
+            _pCPC->_q_events.pop();
         }
-        LeaveCriticalSection(&(pCPC->csEvents));
+        LeaveCriticalSection(&(_pCPC->csEvents));
+
+        if(ev){
+            _pCPC->_p_events->fireEvent(ev->_dispId, &ev->_dispParams);
+            delete ev;
+        }
+        else break;
     }
-out:
-    CoUninitialize();
-    return 0;
+    _HasUnprocessedNotify=false;
 }
 
 ////////////////////////////////////////////////////////////////////////////////////////////////
@@ -345,9 +415,9 @@ VLCDispatchEvent::~VLCDispatchEvent()
 ////////////////////////////////////////////////////////////////////////////////////////////////
 // VLCConnectionPointContainer
 ////////////////////////////////////////////////////////////////////////////////////////////////
-
+extern HMODULE DllGetModule();
 VLCConnectionPointContainer::VLCConnectionPointContainer(VLCPlugin *p_instance) :
-    _p_instance(p_instance), freeze(FALSE), isRunning(TRUE)
+    _p_instance(p_instance), freeze(FALSE), isRunning(TRUE), _ESProxyWnd(0)
 {
     _p_events = new VLCConnectionPoint(dynamic_cast<LPCONNECTIONPOINTCONTAINER>(this),
             _p_instance->getDispEventID());
@@ -361,31 +431,24 @@ VLCConnectionPointContainer::VLCConnectionPointContainer(VLCPlugin *p_instance)
 
     // init protection
     InitializeCriticalSection(&csEvents);
-    ConditionInit(&sEvents);
 
-    // create thread
-    hThread = CreateThread(NULL, NULL, ThreadProcEventHandler,
-                           dynamic_cast<LPVOID>(this), NULL, NULL);
+    _ESProxyWnd = EventSystemProxyWnd::CreateESPWindow(DllGetModule(), this);
 };
 
 VLCConnectionPointContainer::~VLCConnectionPointContainer()
 {
-    EnterCriticalSection(&csEvents);
     isRunning = FALSE;
     freeze = TRUE;
-    ConditionSignal(&sEvents);
-    LeaveCriticalSection(&csEvents);
 
-    do {
-        /* nothing wait for thread to finish */;
-    } while(WaitForSingleObjectEx (hThread, INFINITE, TRUE) == WAIT_IO_COMPLETION);
-    CloseHandle(hThread);
+    if(_ESProxyWnd)
+        _ESProxyWnd->DestroyWindow();
+    _ESProxyWnd=0;
 
-    ConditionDestroy(&sEvents);
     DeleteCriticalSection(&csEvents);
 
     _v_cps.clear();
     while(!_q_events.empty()) {
+        delete _q_events.front();
         _q_events.pop();
     };
 
@@ -435,18 +498,20 @@ void VLCConnectionPointContainer::freezeEvents(BOOL bFreeze)
 
 void VLCConnectionPointContainer::fireEvent(DISPID dispId, DISPPARAMS* pDispParams)
 {
-    EnterCriticalSection(&csEvents);
+    if(_ESProxyWnd){
+        EnterCriticalSection(&csEvents);
 
-    // queue event for later use when container is ready
-    _q_events.push(new VLCDispatchEvent(dispId, *pDispParams));
-    if( _q_events.size() > 1024 )
-    {
-        // too many events in queue, get rid of older one
-        delete _q_events.front();
-        _q_events.pop();
+        // queue event for later use when container is ready
+        _q_events.push(new VLCDispatchEvent(dispId, *pDispParams));
+        if( _q_events.size() > 1024 )
+        {
+            // too many events in queue, get rid of older one
+            delete _q_events.front();
+            _q_events.pop();
+        }
+        LeaveCriticalSection(&csEvents);
+        _ESProxyWnd->NewEventNotify();
     }
-    ConditionSignal(&sEvents);
-    LeaveCriticalSection(&csEvents);
 };
 
 void VLCConnectionPointContainer::firePropChangedEvent(DISPID dispId)
diff --git a/projects/activex/connectioncontainer.h b/projects/activex/connectioncontainer.h
index e4f9589..4f22504 100644
--- a/projects/activex/connectioncontainer.h
+++ b/projects/activex/connectioncontainer.h
@@ -77,7 +77,6 @@ private:
 };
 
 //////////////////////////////////////////////////////////////////////////
-
 class VLCDispatchEvent {
 
 public:
@@ -90,6 +89,7 @@ public:
     DISPPARAMS  _dispParams;
 };
 
+class EventSystemProxyWnd;
 class VLCConnectionPointContainer : public IConnectionPointContainer
 {
 
@@ -126,7 +126,7 @@ public:
 
 public:
     CRITICAL_SECTION csEvents;
-    HANDLE sEvents;
+    EventSystemProxyWnd* _ESProxyWnd;
 
     VLCPlugin *_p_instance;
     BOOL isRunning;
@@ -135,9 +135,6 @@ public:
     VLCConnectionPoint *_p_props;
     std::vector<LPCONNECTIONPOINT> _v_cps;
     std::queue<class VLCDispatchEvent *> _q_events;
-
-private:
-    HANDLE  hThread;
 };
 
 #endif
-- 
1.7.7.1.msysgit.0

